# -*- coding: utf-8 -*-

# Copyright (c) 2018-2022 Fumito Hamamura <fumito.ham@gmail.com>

# This library is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation version 3.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library.  If not, see <http://www.gnu.org/licenses/>.

# The source code contains parts copied and modified from Spyder project:
# https://github.com/spyder-ide/spyder
# See below for the original copyright notice.

#
# Copyright (c) Spyder Project Contributors
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.

import sys
import os
import os.path as osp

import spyder
from spyder.plugins.ipythonconsole.utils.kernelspec import SpyderKernelSpec
from spyder.utils.misc import get_python_executable
from spyder.utils.programs import is_python_interpreter

if spyder.version_info < (5,):
    from spyder.config.manager import CONF

if spyder.version_info > (4, 1):    # 4.1.0 onwards
    from spyder.plugins.ipythonconsole.utils.kernelspec import get_activation_script
    import logging
    logger = logging.getLogger(__name__)
    from spyder.utils.conda import (add_quotes, get_conda_activation_script,
                                    get_conda_env_path, is_conda_env)

if spyder.version_info >= (4, 1, 1) and spyder.version_info <= (5, 5, 1):  # 4.1.1 - 5.5.1
    from spyder.plugins.ipythonconsole.utils.kernelspec import is_different_interpreter


class MxKernelSpec(SpyderKernelSpec):

    CONF_SECTION = 'ipython_console'    # 5.1.0 onwards

    @property
    def argv(self):
        """Command to start kernels"""
        # Python interpreter used to start kernels
        if spyder.version_info < (5, 1):
            if CONF.get('main_interpreter', 'default'):
                pyexec = get_python_executable()
            else:
                if spyder.version_info < (4, 2):
                    # Avoid IPython adding the virtualenv on which Spyder is running
                    # to the kernel sys.path
                    os.environ.pop('VIRTUAL_ENV', None)
                pyexec = CONF.get('main_interpreter', 'executable')
                if not is_python_interpreter(pyexec):
                    pyexec = get_python_executable()
                    CONF.set('main_interpreter', 'executable', '')
                    CONF.set('main_interpreter', 'default', True)
                    CONF.set('main_interpreter', 'custom', False)
        else:
            if self.get_conf('default', section='main_interpreter'):
                pyexec = get_python_executable()
            else:
                pyexec = self.get_conf('executable', section='main_interpreter')
                if not is_python_interpreter(pyexec):
                    pyexec = get_python_executable()
                    self.set_conf('executable', '', section='main_interpreter')
                    self.set_conf('default', True, section='main_interpreter')
                    self.set_conf('custom', False, section='main_interpreter')

        if spyder.version_info <= (4, 1, 1):
            # Fixes spyder-ide/spyder#3427.
            if os.name == 'nt':
                dir_pyexec = osp.dirname(pyexec)
                pyexec_w = osp.join(dir_pyexec, 'pythonw.exe')
                if osp.isfile(pyexec_w):
                    pyexec = pyexec_w

        # Command used to start kernels
        if spyder.version_info < (4, 1):    # 4.0.x
            kernel_cmd = [
                pyexec,
                '-m',
                'spymx_kernels.start_kernel',
                '-f',
                '{connection_file}'
            ]
        else:
            if spyder.version_info == (4, 1, 1):
                is_different = False    # pyexec != sys.executable
            elif spyder.version_info <= (5, 5, 1):
                # Part of spyder-ide/spyder#11819
                is_different = is_different_interpreter(pyexec)
            else:   # Spyder 5.5.3 onward
                is_different = False  # True starts a standard kernel

            if is_different and is_conda_env(pyexec=pyexec):
                # If this is a conda environment we need to call an intermediate
                # activation script to correctly activate the spyder-kernel
                # If changes are needed on this section make sure you also update
                # the activation scripts at spyder/plugins/ipythonconsole/scripts/

                if spyder.version_info < (4, 2):
                    conda_activation_script = get_conda_activation_script()
                else:
                    conda_activation_script = get_conda_activation_script(pyexec)

                kernel_cmd = [
                    get_activation_script(),  # This is bundled with Spyder
                    conda_activation_script,
                    get_conda_env_path(pyexec),  # Might be external
                    pyexec,
                    '{connection_file}',
                ]
            else:
                kernel_cmd = [
                    pyexec,
                    '-m',
                    'spymx_kernels.start_kernel',
                    '-f',
                    '{connection_file}'
                ]
            logger.info('Kernel command: {}'.format(kernel_cmd))

        return kernel_cmd





from qtpy.QtCore import Signal, Slot
from qtpy.QtWidgets import QVBoxLayout
import spyder

from spyder.config.base import _
from spyder.api.preferences import PluginConfigPage
from spyder.api.plugin_registration.decorators import on_plugin_available
from spyder.plugins.ipythonconsole.widgets.main_widget import (
    IPythonConsoleWidgetOptionsMenuSections)

if spyder.version_info > (5, 4, 0):
    from spyder.plugins.ipythonconsole.widgets.main_widget import (
        IPythonConsoleWidgetTabsContextMenuSections
    )

from spyder_modelx.widgets.mxexplorer import MxMainWidget
from .stacked_mixin import MxStackedMixin

# New plugin API since Spyder 5
from spyder.plugins.mainmenu.api import (
    ApplicationMenus, ConsolesMenuSections)
from spyder.api.plugins import SpyderDockablePlugin, Plugins
from qtpy.QtGui import QIcon
from spyder.api.widgets.main_widget import PluginMainWidget


class MxPluginMainWidgetActions:

    OpenNewConsole = 'new_console'
    ConnectToKernel = 'connect_to_kernel'
    SelectInDataView = 'select_in_dataview'
    SelectInNewDataView = 'select_in_new_dataview'


class MxPluginMainWidgetMainToolBarSections:
    Main = 'main_section'


class MxPluginMainWidgetOptionsMenuSections:
    Consoles = 'mx_consoles_section'


class MxConsoleWidgetTabsContextMenuSections:
    Consoles = 'mx_tabs_consoles_section'


class MxPluginMainWidgetBase(MxStackedMixin, PluginMainWidget):

    MX_WIDGET_CLASS = MxMainWidget

    def __init__(self, name=None, plugin=None, parent=None):
        PluginMainWidget.__init__(self, name, plugin, parent)
        MxStackedMixin.__init__(self, parent=parent)

        # Layout
        layout = QVBoxLayout()
        layout.addWidget(self.stack)
        self.setLayout(layout)

        if spyder.version_info > (5, 2):
            self.ipyconsole = plugin.get_plugin(Plugins.IPythonConsole).get_widget()
        else:
            self.ipyconsole = plugin.get_plugin(Plugins.IPythonConsole)

        # To avoid circular dependency,
        # Set by MxAnalyzer and MxDataViewer's register method
        # calling set_child_plugin
        self.analyzer = None
        self.dataview = None

    def set_child_plugin(self, attr, widget):
        setattr(self, attr, widget)

    # --- API: methods to define or override
    # ------------------------------------------------------------------------
    def get_title(self):
        """
        Return the title that will be displayed on dockwidget or window title.
        """
        return _('MxExplorer')

    def setup(self):
        """
        Create widget actions, add to menu and other setup requirements.
        """

        # ---- Toolbar actions
        self.select_dataview_action = select_dataview_action = self.create_action(
            MxPluginMainWidgetActions.SelectInDataView,
            text=_('Select in DataView'),
            icon=self.create_icon('newwindow'),
            triggered=self.select_in_dataview
        )

        self.select_new_dataview_action = select_new_dataview = self.create_action(
            MxPluginMainWidgetActions.SelectInNewDataView,
            text=_('Select in New DataView'),
            icon=self.create_icon('newwindow'),
            triggered=self.select_in_new_dataview
        )

        self.create_client_action = new_console_action = self.create_action(
            MxPluginMainWidgetActions.OpenNewConsole,
            text=_('New MxConsole'),
            icon=self.create_icon('ipython_console'),
            triggered=self.create_new_client
        )

        self.connect_to_kernel_action = self.create_action(
               MxPluginMainWidgetActions.ConnectToKernel,
               text=_("Connect to an existing MxKernel"),
               icon=None,
               tip=_("Open a new IPython console connected to an existing MxKernel"),
               triggered=self.create_client_for_kernel
        )

        # Options menu
        options_menu = self.ipyconsole.get_options_menu()
        for item in [new_console_action, self.connect_to_kernel_action]:
            # Bypass SpyderMenuMixin.add_item_to_menu because of missing before_section parameter.
            options_menu.add_action(
                item,
                section=MxPluginMainWidgetOptionsMenuSections.Consoles,
                before_section=IPythonConsoleWidgetOptionsMenuSections.Consoles
            )

        # Main toolbar
        main_toolbar = self.get_main_toolbar()
        for item in [select_dataview_action, select_new_dataview]:
            self.add_item_to_toolbar(
                item,
                toolbar=main_toolbar,
                section=MxPluginMainWidgetMainToolBarSections.Main,
            )

        if spyder.version_info > (5, 4, 0):
            # Tabs context menu
            tabs_context_menu = self.ipyconsole.tabwidget.menu
            for item in [new_console_action, self.connect_to_kernel_action]:
                tabs_context_menu.add_action(
                    item,
                    section=MxConsoleWidgetTabsContextMenuSections.Consoles,
                    before_section=IPythonConsoleWidgetTabsContextMenuSections.Consoles,
                )

    def update_actions(self):
        """
        Update the state of exposed actions.

        Exposed actions are actions created by the self.create_action method.
        """
        pass

    def select_in_dataview(self):
        self.current_widget().explorer.treeview.select_in_dataview()

    def select_in_new_dataview(self):
        self.current_widget().explorer.treeview.select_in_new_dataview()


class ModelxConfigPage(PluginConfigPage):
    """modelx plugin preferences."""

    def get_name(self):
        return _('modelx')

    def setup_page(self):
        pass


class ModelxPlugin(SpyderDockablePlugin):
    """modelx plugin."""

    NAME = 'modelx_plugin'
    WIDGET_CLASS = None # To be defined in each version
    REQUIRES = [Plugins.IPythonConsole, Plugins.MainMenu]
    CONF_SECTION = 'modelx'
    CONFIGWIDGET_CLASS = ModelxConfigPage
    CONF_FILE = False

    # -------------------------------------------------------------------
    # --- API: Mandatory methods to define ------------------------------

    @staticmethod
    def get_name():
        """
        Return the plugin localized name.

        Returns
        -------
        str
            Localized name of the plugin.

        Notes
        -----
        This is a method to be able to update localization without a restart.
        """
        return _('MxExplorer')

    def get_description(self):
        """
        Return the plugin localized description.

        Returns
        -------
        str
            Localized description of the plugin.

        Notes
        -----
        This is a method to be able to update localization without a restart.
        """
        return _('Main widget of the plugin for modelx')

    def get_icon(self):
        """
        Return the plugin associated icon.

        Returns
        -------
        QIcon
            QIcon instance
        """
        return QIcon()

    def register(self):
        """
        Setup and register plugin in Spyder's main window and connect it to
        other plugins.
        """
        pass

    # register renamed to on_initialize from Spyder 5.1
    def on_initialize(self):
        pass

    @on_plugin_available(plugin=Plugins.IPythonConsole)
    def on_main_menu_available(self):
        widget = self.get_widget()
        mainmenu = self.get_plugin(Plugins.MainMenu)

        # Add signal to update actions state before showing the menu
        console_menu = mainmenu.get_application_menu(
            ApplicationMenus.Consoles)

        for console_new_action in [widget.create_client_action, widget.connect_to_kernel_action]:
            mainmenu.add_item_to_application_menu(
                console_new_action,
                menu_id=ApplicationMenus.Consoles,
                before_section=ConsolesMenuSections.New,
            )




from collections import namedtuple
from qtpy.QtCore import QAbstractItemModel, QModelIndex, Qt, Slot, QMetaMethod
from qtpy.QtWidgets import (
    QTableView, QApplication, QTreeView, QStyledItemDelegate,
    QMessageBox, QLineEdit, QWidget, QVBoxLayout, QScrollArea,
    QDialogButtonBox, QPushButton, QHBoxLayout, QLabel, QSplitter
)

COLS = COL_TITLE, COL_VALUE = range(2)
COL_HEADER = "Property", "Value"

# ------ ListItem classes -----------------------------------------------------


class BaseListItem:

    child_key = None
    child_class = None

    def __init__(self, parent, index, data):

        self.parent = parent
        self.index = index
        self.child_items = []
        self.attrs = {}

        for key, val in data.items():
            if key == self.child_key:
                for i, child_data in enumerate(val):
                    self.child_items.append(
                        self.child_class(self, i, child_data))
            else:
                self.attrs[key] = val

    def get_display_role(self, column):
        raise NotImplementedError


class ValueListItem(BaseListItem):

    child_key = "values"

    def get_display_role(self, column):
        return None


class ValueItem(BaseListItem):

    child_key = "refs"

    def get_display_role(self, column):
        if column == 0:
            return self.attrs["value"]
        elif column == 1:
            spec = self.attrs["spec"]
            return spec["path"] if spec else None
        else:
            return None


class RefItem(BaseListItem):

    child_key = None
    child_class = None

    def get_display_role(self, column):
        if column == 0:        # ref name with parents
            return ".".join(self.attrs["fullname"].split(".")[1:])
        else:
            return None


ValueListItem.child_class = ValueItem
ValueItem.child_class = RefItem


# ------ MxDataList Model and View --------------------------------------------


class MxDataListView(QTreeView):

    def __init__(self, parent=None):
        QTreeView.__init__(self, parent)
        self._parent = parent
        # self.formulaPane = parent.formulaPane
        # self.setRootIsDecorated(False)
        # self.setAlternatingRowColors(False)
        # self.doubleClicked.connect(self.openEditor)
        # self.setItemDelegate(MxPropertyItemDelegate(self))
        # self.setContentsMargins(0, 0, 0, 0)

    def currentChanged(self, current: QModelIndex, previous: QModelIndex) -> None:
        if current.isValid():
            item = current.internalPointer()
            if isinstance(item, ValueItem) and item.attrs["spec"]:
                self.parent().update_dataattrs(item)
            else:
                self.parent().update_dataattrs(None)


class MxDataListModel(QAbstractItemModel):
    """
        value      | spec.path      # 1st level
         |
         +-ref                      # 2nd level
    """

    def __init__(self, parent=None, data=None):
        super(MxDataListModel, self).__init__(parent)
        self.resetDataList(data)

    def resetDataList(self, data):
        self.beginResetModel()
        self.value_info = ValueListItem(
            parent=None,
            index=0,
            data={"values": data}
        )
        self.endResetModel()

    def _get_item(self, index):

        if index.isValid():
            return index.internalPointer()
        else:
            return self.value_info

    def rowCount(self, parent):     # pure virtual
        return len(self._get_item(parent).child_items)

    def columnCount(self, parent):             # pure virtual
        return 2

    def index(self, row, column, parent):      # pure virtual

        if self.hasIndex(row, column, parent):
            parent_item = self._get_item(parent)
            return self.createIndex(row, column, parent_item.child_items[row])

        else:
            return QModelIndex()

    def parent(self, index):        # pure virtual

        parent_item = self._get_item(index).parent

        if isinstance(parent_item, ValueListItem):
            return QModelIndex()
        else:
            return self.createIndex(parent_item.index, 0, parent_item)

    def data(self, index, role):    # pure virtual

        if role != Qt.DisplayRole:
            return None

        row = index.row()
        col = index.column()

        item = self._get_item(index)
        return item.get_display_role(col)

    def flags(self, index):                                # virtual
        if not index.isValid():
            return Qt.NoItemFlags

        return Qt.ItemIsEnabled | Qt.ItemIsSelectable

    def headerData(self, column, orientation, role):       # virtual

        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            if column == 0:
                return "Value/Name"
            else:
                return "Path"
        return None

# ------ Main widget in the tab -----------------------------------------------


class MxDataListWidget(QSplitter):

    def __init__(self, parent, orientation):
        super().__init__(parent, orientation=orientation)
        self.shell = None   # To be set by MxShellWidget
        self.plugin = parent.plugin
        self._parent = parent
        self.datalist = MxDataListView(self)
        self.dataattrs = MxDataAttrsView(self)

    def process_remote_view(self, data):
        if data:
            model = self.datalist.model()
            if model:
                model.resetDataList(data)
            else:
                self.datalist.setModel(
                    MxDataListModel(parent=self, data=data))
        else:
            self.datalist.setModel(None)

    def update_dataattrs(self, item):
        if item:
            model = self.dataattrs.model()
            if model:
                model.resetItemData(item)
            else:
                self.dataattrs.setModel(
                    MxDataAttrsModel(parent=self, item=item))
        else:
            self.dataattrs.setModel(None)

# ------ MxDataAttrs Model and View -------------------------------------------


class MxDataAttrsView(QTreeView):

    def __init__(self, parent=None):
        QTreeView.__init__(self, parent)
        self._parent = parent
        # self.formulaPane = parent.formulaPane
        self.setRootIsDecorated(False)
        self.setAlternatingRowColors(False)
        # self.doubleClicked.connect(self.openEditor)
        # self.setItemDelegate(MxPropertyItemDelegate(self))
        self.setContentsMargins(0, 0, 0, 0)


class MxDataAttrsModel(QAbstractItemModel):

    def __init__(self, parent=None, item=None):
        super(MxDataAttrsModel, self).__init__(parent)
        self.resetItemData(item)

    def resetItemData(self, item):
        self.beginResetModel()
        self.spec = {}
        for k, v in item.attrs["spec"].items():
            if isinstance(v, str):
                self.spec[k] = v
        self.endResetModel()

    def rowCount(self, parent):     # pure virtual
        if not parent.isValid():
            return len(self.spec)
        else:
            return 0

    def columnCount(self, parent):  # pure virtual
        return 2

    def index(self, row, column, parent):      # pure virtual
        if self.hasIndex(row, column, parent):
            return self.createIndex(row, column, self.spec)
        else:
            return QModelIndex()

    def parent(self, index):        # pure virtual
        return QModelIndex()

    def data(self, index, role):    # pure virtual
        if not index.isValid():
            return None

        if role != Qt.DisplayRole:
            return None

        row = index.row()
        col = index.column()

        for i, key in enumerate(self.spec):
            if i == row:
                break

        if col == 0:
            return key
        elif col == 1:
            return self.spec[key]
        else:
            raise ValueError("must not happen")

    def flags(self, index):                                # virtual
        if not index.isValid():
            return Qt.NoItemFlags

        return Qt.ItemIsEnabled | Qt.ItemIsSelectable

    def headerData(self, column, orientation, role):       # virtual

        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            if column == 0:
                return "Parameter"
            elif column == 1:
                return "Value"

        return None




